#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "olaf_fp_extractor.h"
#include "olaf_config.h"


struct Olaf_FP_Extractor{
	int audioBlockIndex;
	struct extracted_fingerprints fingerprints;
	uint32_t * recentFingerprints;
	int recentFingerprintIndex;
	Olaf_Config * config;
};

Olaf_FP_Extractor * olaf_fp_extractor_new(Olaf_Config * config){

	Olaf_FP_Extractor *fp_extractor = (Olaf_FP_Extractor*) malloc(sizeof(Olaf_FP_Extractor));

	fp_extractor->config = config;

	fp_extractor->recentFingerprintIndex = 0;

	fp_extractor->recentFingerprints = (uint32_t *) malloc(config->recentFingerprintSize * sizeof(uint32_t));

	fp_extractor->fingerprints.fingerprints = (struct fingerprint*) malloc(config->maxFingerprints * sizeof(struct fingerprint));

	fp_extractor->fingerprints.fingerprintIndex = 0;

	return fp_extractor;
}

void olaf_fp_extractor_destroy(Olaf_FP_Extractor * fp_extractor){
	free(fp_extractor->fingerprints.fingerprints);
	free(fp_extractor->recentFingerprints);
	free(fp_extractor);
}

int compareEventPoints (const void * a, const void * b) {
	struct eventpoint aPoint = *(struct eventpoint*) a;
	struct eventpoint bPoint = *(struct eventpoint*) b;
	return (aPoint.timeIndex - bPoint.timeIndex);
}

// Calculates a hash from a fingerprint struct. A hash fits in a 32bit unsigned int.
//
// The hash consists of time info and time independent info.
// The LSB's of the hash are used to store the time information.
// This makes it so that when the hashes are stored __in order__ matches can be 
// found using the MSB's.
//
// An example: say that the MSB part is 1546 and the time info 11, 
// combined this makes 154611
// The reference database contains for example:
//
// ...
// 1000 17
// 1264 20
// 1164 01
// 1164 07
// 1546 12 <--
// 1546 14 <--
// 1546 19 <--
// 1546 77 <--
// 1687 16
// ...
//
//A binary search for 1546xx in the reference database ends up at one of the above arrows. 
//Subsequently all the 1546xx values are identified by iteration in both directions.
//
uint32_t olaf_fp_extractor_hash(struct fingerprint f){
	uint16_t frequency1 = f.frequencyBin1;
	uint16_t deltaF = abs(f.frequencyBin1 - f.frequencyBin2);
	uint16_t deltaT = f.timeIndex1 - f.timeIndex2;
	uint16_t f1LargerThanF2 = f.frequencyBin1 > f.frequencyBin2 ? 1 : 0;
	
	//uint16_t frequency1Interpolated = (int) (f.fractionalFrequencyBin1 * 2);
	//                                    21 bits in total	
	// frequency 0-255    -  8 bits    -  13
	// deltaF    0-63     -  6 bits    -   7
	// f1<f2     boolean  -  1 bits    -   6
	// deltaT    0-63     -  6 bits    -   0   about 2 seconds
	
	uint32_t fp_hash = ((frequency1     &  ((1<<10) -1)  ) << 13) +
	                   ((deltaF         &  ((1<<6) -1)   ) <<  7) +
	                   ((f1LargerThanF2 &  ((1<<1) -1)   ) <<  6) +
	                   ((deltaT         &  ((1<<6) -1)   ) <<  0) ;

	return fp_hash;
}

struct extracted_fingerprints * olaf_fp_extractor_extract(Olaf_FP_Extractor * fp_extractor,struct extracted_event_points * eventPoints,int audioBlockIndex){
	for(int i = 0; i < eventPoints->eventPointIndex ; i++){

		int t1 = eventPoints->eventPoints[i].timeIndex;
		int f1 = eventPoints->eventPoints[i].frequencyBin;
		int ppp1 = eventPoints->eventPoints[i].printsPerPoint;

		//do not evaluate empty points
		if(f1==0 && t1==0)
			break;

		//do not evaulate points with more than x prints per event point
		if(ppp1>fp_extractor->config->maxFingerprintsPerPoint)
			continue;

		//do not evaluate event points that are too recent
		int diffToCurrentTime = audioBlockIndex-fp_extractor->config->maxTimeDistance;
		if(t1>diffToCurrentTime)
			break;

		for(int j = i+1; j < eventPoints->eventPointIndex; j++){

			int t2 =  eventPoints->eventPoints[j].timeIndex;
			int f2 =  eventPoints->eventPoints[j].frequencyBin;
			int ppp2 =  eventPoints->eventPoints[j].printsPerPoint;

			int fDiff = abs(f1 - f2);
			int tDiff = t2-t1;

			assert(t2>=t1);

			//do not evaluate points to far in the future
			if(tDiff > fp_extractor->config->maxTimeDistance)
				break;

			//do not evaulate points with more than x prints per event point
			if(ppp2>fp_extractor->config->maxFingerprintsPerPoint)
				continue;

			if(tDiff >= fp_extractor->config->minTimeDistance && 
				tDiff <= fp_extractor->config->maxTimeDistance && 
				fDiff >= fp_extractor->config->minFreqDistance && 
				fDiff <= fp_extractor->config->maxFreqDistance){

				assert(t2>t1);
				
				//temporarily store (do not increment fingerprint index, unless it is not yet discovered)
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].timeIndex1 = t1;
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].timeIndex2 = t2;
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].frequencyBin1 = f1;
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].fractionalFrequencyBin1 = eventPoints->eventPoints[i].fractionalFrequencyBin;
				
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].frequencyBin2 = f2;
				fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex].fractionalFrequencyBin2 = eventPoints->eventPoints[j].fractionalFrequencyBin;
				

				uint32_t fingerprintHash = olaf_fp_extractor_hash(fp_extractor->fingerprints.fingerprints[fp_extractor->fingerprints.fingerprintIndex]);

				//check if it is already discovered
				bool newFingerprint = true;
				for(int k = 0 ; k < fp_extractor->config->recentFingerprintSize ; k++ ){
					if(fingerprintHash == fp_extractor->recentFingerprints[k]){
						newFingerprint = false;
						break;
					}
				}

				if(newFingerprint){
					eventPoints->eventPoints[i].printsPerPoint++;
					eventPoints->eventPoints[j].printsPerPoint++;

					//int fDiffBins = abs(f1 - f2);
					//fprintf(stderr,"t1: %d  t2: %d  tdiff: %d (bins) f1: %d f2: %d  fdiff: %d (cents) fdiff: %d (bins)\n",t1,t2,tDiff,f1,f2,fDiff,fDiffBins);
					
					fp_extractor->fingerprints.fingerprintIndex++;

					fp_extractor->recentFingerprints[fp_extractor->recentFingerprintIndex] = fingerprintHash;
					fp_extractor->recentFingerprintIndex++;
					if(fp_extractor->recentFingerprintIndex == fp_extractor->config->recentFingerprintSize)
						fp_extractor->recentFingerprintIndex=0;

				}
			}
		}
	}

	int cutoffTime = eventPoints->eventPoints[eventPoints->eventPointIndex-1].timeIndex - fp_extractor->config->maxTimeDistance;

	//prepare the array for the next event loops
	for(int i = 0 ; i < eventPoints->eventPointIndex ; i++){

		//mark the event points that are too old
		if(eventPoints->eventPoints[i].timeIndex <= cutoffTime){
			eventPoints->eventPoints[i].timeIndex = 1<<16;
			eventPoints->eventPoints[i].printsPerPoint = 0;
			eventPoints->eventPoints[i].frequencyBin = 0;
			eventPoints->eventPoints[i].magnitude = 0;
		}

		//mark the event points that are used the max number of times
		if(eventPoints->eventPoints[i].printsPerPoint > fp_extractor->config->maxFingerprintsPerPoint){
			eventPoints->eventPoints[i].timeIndex = 1<<16;
			eventPoints->eventPoints[i].printsPerPoint = 0;
			eventPoints->eventPoints[i].frequencyBin = 0;
		}
	}

	//sort the array from low timeIndex to high
	//the marked event points have a high time index
	qsort(eventPoints->eventPoints,eventPoints->eventPointIndex, sizeof(struct eventpoint), compareEventPoints);

	//find the first marked event point: this is where the next point needs to be stored
	for(int i = 0 ; i <eventPoints->eventPointIndex ; i++){

		if(eventPoints->eventPoints[i].timeIndex==1<<16){
			//the next point needs to be stored at this index
			//fprintf(stderr,"New event point index %d , old index %d,  current time %d (%d ms) , cutoff time %d \n",i,eventPoints->eventPointIndex,audioBlockIndex,audioBlockIndex * 32,cutoffTime);
			eventPoints->eventPointIndex  = i;

			//for(int j = 0 ; j < eventPoints->eventPointIndex ; j++){
			//	fprintf(stderr,"\tNew ev %d %d %d\n",eventPoints->eventPoints[j].timeIndex,eventPoints->eventPoints[j].frequencyBin,eventPoints->eventPoints[j].printsPerPoint); 
			//}

			break;
		}
	}
	//fprintf(stderr,"New ev\n");

	return &fp_extractor->fingerprints;
}
